/* -*- add-client.c: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
/*
 * Licensed under the Academic Free License version 2.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> 
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <dbus/dbus.h>

#define NUM_THREADS 20
#define CALLS_PER_THREAD 10000

const char *const INTERFACE_NAME = "com.test.dbus_example";
const char *const SERVER_BUS_NAME = "com.test.add_server";
const char *const CLIENT_BUS_NAME = "com.test.add_client";
const char *const SERVER_OBJECT_PATH_NAME = "/com/test/adder";
const char *const CLIENT_OBJECT_PATH_NAME = "/com/test/add_client";
const char *const METHOD_NAME = "add_numbers";

DBusError dbus_error;
void print_dbus_error (char *str);
DBusConnection *conn;
int threads_alive = 0;
pthread_mutex_t lock;

dbus_bool_t addTimeout(DBusTimeout *timeout, void* data){
	printf("Thread = %ld, Adding timeout = %p\n", syscall(__NR_gettid),timeout);
}

void removeTimeout(DBusTimeout *timeout, void* data){
	printf("Thread = %ld,Removing timeout = %p\n", syscall(__NR_gettid),timeout);
}

void toggleTimeout(DBusTimeout *timeout, void* data){
	printf("Thread = %ld,Removing timeout = %p\n", syscall(__NR_gettid),timeout);
}


void *thread_func(void *vargp)
{
	int i = CALLS_PER_THREAD;
	threads_alive++;
	while(i > 0) {
		usleep(10*1000);

		DBusMessage *request;

		if ((request = dbus_message_new_method_call (SERVER_BUS_NAME, SERVER_OBJECT_PATH_NAME,
				INTERFACE_NAME, METHOD_NAME)) == NULL) {
			fprintf (stderr, "Error in dbus_message_new_method_call\n");
			exit (1);
		}

		DBusMessageIter iter;
		dbus_message_iter_init_append (request, &iter);
		char *ptr = "1 4";
		if (!dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &ptr)) {
			fprintf (stderr, "Error in dbus_message_iter_append_basic\n");
			exit (1);
		}
		//		DBusPendingCall *pending_return;
		DBusError error;
		memset(&error,0,sizeof(error));


//		pthread_mutex_lock(&lock);
		DBusMessage *reply = dbus_connection_send_with_reply_and_block(conn, request, 25000, &error);
//		pthread_mutex_unlock(&lock);

		dbus_uint32_t m_id = dbus_message_get_serial(request);
		printf("Reply for %u\n",m_id);

		if (reply != NULL) {

			dbus_message_unref (request);
			char *s;
			if (dbus_message_get_args (reply, &dbus_error, DBUS_TYPE_STRING, &s, DBUS_TYPE_INVALID)) {
				//				printf ("%s\n", s);
			}
			else
			{
				fprintf (stderr, "Did not get arguments in reply\n");
				exit (1);
			}
			dbus_message_unref (reply);

		} else {
			fprintf (stderr, "Error in dbus_connection_send_with_reply %s\n",error.message);
		}

		i--;
	}
	threads_alive--;
	return NULL;
}

int main (int argc, char **argv)
{

	int ret;
	if (pthread_mutex_init(&lock, NULL) != 0)
	{
		printf("\n mutex init has failed\n");
		return 1;
	}
	dbus_error_init (&dbus_error);

	conn = dbus_bus_get (DBUS_BUS_SESSION, &dbus_error);

	if (dbus_error_is_set (&dbus_error))
		print_dbus_error ("dbus_bus_get");

	if (!conn)
		exit (1);

	// Get a well known name
	while (1) {
		ret = dbus_bus_request_name (conn, CLIENT_BUS_NAME, 0, &dbus_error);

		if (ret == DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
			break;

		if (ret == DBUS_REQUEST_NAME_REPLY_IN_QUEUE) {
			fprintf (stderr, "Waiting for the bus ... \n");
			sleep (1);
			continue;
		}
		if (dbus_error_is_set (&dbus_error))
			print_dbus_error ("dbus_bus_get");
	}

	//	if (!dbus_connection_set_timeout_functions(conn, addTimeout, removeTimeout,
	//			toggleTimeout, NULL , NULL))
	//	{
	//		printf("Timeout func registration failed\n");
	//		exit(1);
	//	}

	for(int i = 0 ; i < NUM_THREADS; i++) {
		pthread_t tid;
		pthread_create(&tid, NULL, thread_func, NULL);
	}

	while(threads_alive > 0) {
		//		usleep(10*1000);
		if(dbus_connection_get_dispatch_status(conn)
				== DBUS_DISPATCH_DATA_REMAINS) {
			dbus_connection_read_write_dispatch(conn, 100);
		}
	}

	if (dbus_bus_release_name (conn, CLIENT_BUS_NAME, &dbus_error) == -1) {
		fprintf (stderr, "Error in dbus_bus_release_name\n");
		exit (1);
	}
    pthread_mutex_destroy(&lock);

	return 0;
}

void print_dbus_error (char *str) 
{
	fprintf (stderr, "%s: %s\n", str, dbus_error.message);
	dbus_error_free (&dbus_error);
}
